import os
import random
import sys

import anndata as ad
import numpy as np
import pandas as pd
import torch
import torch.nn.modules.loss
from scipy import sparse
from sklearn.preprocessing import normalize
from torch import Tensor, nn
from torch.nn.functional import binary_cross_entropy, mse_loss
from torch.utils.data import DataLoader
from torch_geometric.data import Data
from torch_geometric.nn import VGAE, GCNConv, InnerProductDecoder
from tqdm import tqdm

from TRAPT.Tools import RPMatrix


def seed_torch(seed=23):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


class InnerProductDecoderWeight(InnerProductDecoder):
    def __init__(self, A_e, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.A_e = A_e

    def forward(
        self, z: Tensor, edge_index: Tensor = None, sigmoid: bool = True
    ) -> Tensor:
        adj = torch.matmul(z, z.t())
        return torch.sigmoid(adj) if sigmoid else adj


class VariationalGCNEncoder(torch.nn.Module):
    def __init__(self, in_channels, h_dim, z_dim):
        super(VariationalGCNEncoder, self).__init__()
        self.conv_h = GCNConv(in_channels, h_dim, cached=True)
        self.conv_mu = GCNConv(h_dim, z_dim, cached=True)
        self.conv_logstd = GCNConv(h_dim, z_dim, cached=True)

    def predict_h(self, x, edge_index):
        h = self.conv_h(x, edge_index).relu()
        return h

    def forward(self, x, edge_index):
        h = self.conv_h(x, edge_index).relu()
        return self.conv_mu(h, edge_index), self.conv_logstd(h, edge_index)


class CVAE(nn.Module):
    def __init__(self, input_dim, condition_dim, h_dim, z_dim):
        super(CVAE, self).__init__()
        self.h_encoder = torch.nn.Linear(input_dim + condition_dim, h_dim)
        self.act = torch.nn.ReLU()
        self.dropout = torch.nn.Dropout()
        self.layer_mu = torch.nn.Linear(h_dim, z_dim)
        self.layer_logstd = torch.nn.Linear(h_dim, z_dim)
        self.decoder = torch.nn.Linear(z_dim + condition_dim, input_dim)

    def reparametrize(self, mu: Tensor, logstd: Tensor) -> Tensor:
        if self.training:
            return mu + torch.randn_like(logstd) * torch.exp(logstd)
        else:
            return mu

    def predict_h(self, x):
        h = self.act(self.h_encoder(x))
        return h

    def kl_div(self):
        return torch.mean(
            0.5
            * torch.sum(
                torch.exp(self.logstd) + self.mu**2 - 1.0 - self.logstd, dim=1
            )
        )

    def forward(self, x):
        condition = x[:, -2:]
        h = self.act(self.h_encoder(x))
        self.mu = self.layer_mu(h)
        self.logstd = self.layer_logstd(h)
        z = self.reparametrize(self.mu, self.logstd)
        decoded = self.decoder(torch.concat([z, condition], dim=1))
        return decoded

class CalcSTM:
    def __init__(self, RP_Matrix, type, checkpoint_path, device='cpu'):
        self.type = type
        assert type in ['H3K27ac', 'ATAC']
        self.checkpoint_path = checkpoint_path
        self.device = device
        print(f"Using {self.device}")
        self.RP_Matrix = RP_Matrix
        self.data = torch.concat(
            [
                torch.tensor(self.RP_Matrix.TR.to_df().values),
                torch.tensor(self.RP_Matrix.Sample.to_df().values),
            ],
            dim=0,
        ).to(self.device)

    @staticmethod
    def get_cos_similar_matrix(m1, m2):
        num = m1.dot(m2.T)
        denom = np.linalg.norm(m1, axis=1).reshape(-1, 1) * np.linalg.norm(m2, axis=1)
        res = num / denom
        res[np.isneginf(res)] = 0
        return res

    def get_edge_index(self, A, B, n=10):
        self.A_s = 0
        self.A_e = A.shape[0]
        self.B_s = self.A_e
        self.B_e = self.B_s + B.shape[0]
        adj = np.zeros((self.B_e, self.B_e))
        AB_cos_sim = self.get_cos_similar_matrix(A.to_df().values, B.to_df().values)
        AB = (AB_cos_sim >= np.sort(AB_cos_sim, axis=1)[:, -n].reshape(-1, 1)).astype(
            int
        )
        adj[self.A_s : self.A_e, self.B_s : self.B_e] = AB
        adj[self.B_s : self.B_e, self.A_s : self.A_e] = AB.T
        row, col, value = sparse.find(adj)
        return torch.tensor(np.array([row, col]), dtype=torch.long)

    @staticmethod
    def sparse_to_tensor(data, type='sparse'):
        if type == 'dense':
            return torch.tensor(data.to_df().values)
        else:
            row, col, value = sparse.find(data)
            indices = torch.tensor(np.array([row, col]), dtype=torch.long)
            values = torch.tensor(value, dtype=torch.float)
            size = torch.Size(data.shape)
            return torch.sparse_coo_tensor(indices, values, size)

    def save_graph(self):
        x, edge_index = self.data.x.to(self.device), self.data.edge_index.to(
            self.device
        )
        z = self.model_vgae.encode(x, edge_index).cpu().detach().numpy()
        A_pred = z.dot(z.T)
        TR_info = pd.DataFrame(list(self.RP_Matrix.TR.obs_names))
        TR_info['type'] = 'TR'
        Sample_info = pd.DataFrame(list(self.RP_Matrix.Sample.obs_names))
        Sample_info['type'] = self.type
        info = pd.concat([TR_info, Sample_info]).astype(str)
        info.columns = ['index', 'type']
        info = info.set_index('index')
        data = ad.AnnData(A_pred, obs=info, var=info)
        data.write_h5ad(f'{self.checkpoint_path}/A_pred_{self.type}.h5ad')

    def recon_loss(self, z):
        y_pred = self.model_vgae.decoder(z, sigmoid=True)
        loss = (
            self.norm
            * binary_cross_entropy(y_pred.view(-1), self.y, weight=self.weight).mean()
        )
        return loss

    def init_vgae(self, h, use_kd):
        seed_torch()
        self.h = h
        self.edge_index = self.get_edge_index(self.RP_Matrix.TR, self.RP_Matrix.Sample)
        self.data = Data(x=self.data, edge_index=self.edge_index)
        x = self.data.x.to(self.device)
        edge_index = self.data.edge_index.to(self.device)
        self.num_features = self.data.num_features
        self.num_nodes = self.data.num_nodes
        values = torch.ones(self.edge_index.shape[1])
        size = torch.Size([self.num_nodes] * 2)
        y = torch.sparse_coo_tensor(self.edge_index, values, size)
        y = y.to_dense().view(-1)
        pos_weight = (self.num_nodes**2 - y.sum()) / y.sum()
        weight = torch.ones_like(y)
        weight[y.numpy().astype(bool)] = pos_weight
        self.weight = weight.to(self.device)
        self.norm = self.num_nodes**2 / ((self.num_nodes**2 - y.sum()) * 2)
        self.y = y.to(self.device)

        encoder = VariationalGCNEncoder(self.data.num_features, self.h_dim, self.z_dim)
        decoder = InnerProductDecoderWeight(self.A_e)
        self.model_vgae = VGAE(encoder, decoder).to(self.device)
        self.optimizer_vgae = torch.optim.Adam(self.model_vgae.parameters(), lr=0.01)
        if os.path.exists(f'{self.checkpoint_path}/model_student_{self.type}.pt'):
            checkpoint = torch.load(
                f'{self.checkpoint_path}/model_student_{self.type}.pt',
                map_location=self.device,
            )
            self.model_vgae.load_state_dict(checkpoint['model_state_dict'])
            self.optimizer_vgae.load_state_dict(checkpoint['optimizer_state_dict'])
        else:
            print('# Training ...')
            for epoch in range(self.epochs_vgae):
                self.model_vgae.train()
                self.optimizer_vgae.zero_grad()
                h = self.model_vgae.encoder.predict_h(x, edge_index)
                z = self.model_vgae.encode(x, edge_index)
                if use_kd:
                    kd_loss = mse_loss(h, self.h)
                else:
                    kd_loss = 0
                re_loss = self.recon_loss(z)
                kl_loss = (1 / self.num_nodes) * self.model_vgae.kl_loss()
                loss = re_loss + kl_loss + kd_loss
                loss.backward()
                self.optimizer_vgae.step()
                if epoch % 1 == 0:
                    print(
                        f'epoch: {epoch}, loss: {loss}, kl_loss: {kl_loss}, re_loss: {re_loss}, kd_loss: {kd_loss}'
                    )
                    checkpoint = {
                        'model_state_dict': self.model_vgae.state_dict(),
                        'optimizer_state_dict': self.optimizer_vgae.state_dict(),
                    }
                    torch.save(
                        checkpoint,
                        f'{self.checkpoint_path}/model_student_{self.type}.pt',
                    )
            torch.save(
                checkpoint,
                f'{self.checkpoint_path}/model_student_{self.type}.pt',
            )
        self.save_graph()

    def init_cvae(self):
        seed_torch()
        condition = np.ones(self.data.shape[0])
        condition[: self.RP_Matrix.TR.shape[0]] = 0
        condition = np.eye(2)[condition.astype(int)]
        condition = normalize(condition, 'l1', axis=0)
        condition = torch.tensor(condition, dtype=torch.float32).to(self.device)
        dataset = torch.concat([self.data, condition], dim=1)
        batch_size = 32
        train_loader = DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=True,
        )
        self.model_cvae = CVAE(
            self.data.shape[1], condition.shape[1], self.h_dim, self.z_dim
        ).to(self.device)
        self.optimizer_cvae = torch.optim.Adam(self.model_cvae.parameters(), lr=0.01)
        loss_func = nn.MSELoss()
        if os.path.exists(f'{self.checkpoint_path}/model_teacher_{self.type}.pt'):
            checkpoint = torch.load(
                f'{self.checkpoint_path}/model_teacher_{self.type}.pt',
                map_location=self.device,
            )
            self.model_cvae.load_state_dict(checkpoint['model_state_dict'])
            self.optimizer_cvae.load_state_dict(checkpoint['optimizer_state_dict'])
        else:
            print('# Training ...')
            for epoch in range(self.epochs_cvae):
                for batch in tqdm(train_loader):
                    self.model_cvae.train()
                    self.optimizer_cvae.zero_grad()
                    decoded = self.model_cvae(batch)
                    kl_loss = self.model_cvae.kl_div() / batch_size
                    re_loss = (
                        loss_func(decoded.view(-1), batch[:, :-2].reshape(-1))
                        / batch_size
                    )
                    loss = re_loss + kl_loss
                    loss.backward()
                    self.optimizer_cvae.step()
                print(
                    f'epoch: {epoch}, loss: {loss}, kl_loss: {kl_loss}, re_loss: {re_loss}'
                )
                checkpoint = {
                    'model_state_dict': self.model_cvae.state_dict(),
                    'optimizer_state_dict': self.optimizer_cvae.state_dict(),
                }
                torch.save(
                    checkpoint,
                    f'{self.checkpoint_path}/model_teacher_{self.type}.pt',
                )

        return self.model_cvae.predict_h(dataset).detach()

    def run(self,use_kd=True):
        self.epochs_cvae = 10
        self.epochs_vgae = 100
        self.h_dim = 48
        self.z_dim = 24
        # Teacher
        z = self.init_cvae()
        # Student
        self.init_vgae(z,use_kd)


if __name__ == '__main__':
    type = sys.argv[1]
    library = sys.argv[2]
    use_kd = True
    checkpoint_path = library
    if len(sys.argv) > 3:
        use_kd = False if sys.argv[3].lower() == "false" else True
    if len(sys.argv) > 4:
        checkpoint_path = sys.argv[4]

    assert type in ['H3K27ac', 'ATAC']

    class RP_Matrix:
        TR = RPMatrix(library, 'RP_Matrix_TR.h5ad').norm('l1').get_data()
        Sample = RPMatrix(library, f'RP_Matrix_{type}.h5ad').norm('l1').get_data()

    CalcSTM(RP_Matrix, type, checkpoint_path).run(use_kd)



import random

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras import backend as K
from keras import Input, Model
from keras.layers import Dense, Dropout, Layer
from keras.models import Model
from keras.optimizers import Adam
from keras.losses import BCE,MSE
from keras.activations import relu 
from keras.regularizers import Regularizer,L2
from keras.callbacks import EarlyStopping

def seed_tensorflow(seed=2023):
    random.seed(seed)
    np.random.seed(seed)
    tf.random.set_seed(seed)

class CustomSigmoid(Layer):
    def __init__(self, *, activity_regularizer=None, trainable=True, dtype=None, autocast=True, name=None, t=1, **kwargs):
        super().__init__(activity_regularizer=activity_regularizer, trainable=trainable, dtype=dtype, autocast=autocast, name=name, **kwargs)
        self.t = t
    def call(self, x):
        return 1 / (1 + K.exp(-x / self.t))

class SparseGroupLasso(Regularizer):
    """
    https://keras-ssg-lasso.readthedocs.io/en/latest/
    argmin(w) (1/2n * ||y - Xw||^2 + λ1 * ||w||1 + λ2 * ∑g∈G √pg * ||w_g||2)
    """

    def __init__(self, l1=0.01, l2=0.01, groups=None):
        self.l1 = l1
        self.l2 = l2
        groups = np.array(groups).astype("int32")
        self.p_l = K.variable(np.sqrt(np.bincount(groups)).reshape((1, -1)))
        self.groups = K.variable(groups, "int32")

    def __call__(self, x):
        regularization = 0.0
        if self.l1:
            regularization += self.l1 * K.sum(K.abs(x))
        if self.l2:
            regularization += self.l2 * K.sum(
                K.dot(
                    self.p_l,
                    K.sqrt(
                        tf.math.unsorted_segment_sum(
                            K.square(x), self.groups, num_segments=self.p_l.shape[1]
                        )
                    ),
                )
            )
        return regularization

    def get_config(self):
        return {
            "name": self.__class__.__name__,
            "l1": self.l1,
            "l2": self.l2,
            "groups": self.groups,
        }


class FeatureSelection:
    def __init__(self, args, data_ad, type):
        self.args = args
        self.data_ad = data_ad
        self.type = type
        self.epochs = 16
        self.batch_size = 32
        self.learning_rate = 0.001
        self.group_size = 10
        self.top = 10
        self.shuffle = True
        self.early_stopping = EarlyStopping(patience=4)

    def get_loss(self):
        if self.args.liner:
            return BCE
        else:
            return MSE
        
    def get_act(self,t=1):
        if self.args.liner:
            return relu
        else:
            return CustomSigmoid(t=t)

    def TSFS(self, X, T):
        seed_tensorflow()
        pos_weight = float(len(T) - T.sum()) / T.sum()
        sample_weight = np.ones(len(T))
        sample_weight[T.astype(bool)] = pos_weight
        if self.args.use_kd:
            input = Input(X.shape[1:])
            y = Dense(
                2, 
                activation="relu"
            )(input)
            feature_extraction = Model(input, y)
            t = feature_extraction(input)
            output = Dense(1)(t)
            output = self.get_act()(output)
            teacher = Model(input, output)
            teacher.compile(
                optimizer=Adam(self.learning_rate),
                loss=self.get_loss(), 
                weighted_metrics=[]
            )
            print("U-RP teacher model running...")
            teacher.fit(
                X,
                T,
                epochs=self.epochs,
                batch_size=self.batch_size,
                sample_weight=sample_weight,
                shuffle=self.shuffle,
                callbacks=[self.early_stopping],
                validation_split=0.15,
                verbose=0,
            )
            Y = feature_extraction.predict(X)
            seed_tensorflow()
            input = Input(X.shape[1:])
            output = Dense(
                Y.shape[-1] * 10,
                activation="relu",
                kernel_regularizer=SparseGroupLasso(groups=self.groups)
            )(input)
            output = Dense(Y.shape[-1], activation="relu")(output)
            fs_model = Model(input, output)
            fs_model.compile(optimizer=Adam(self.learning_rate), loss=MSE, weighted_metrics=[])
            print("U-RP student model running...")
            fs_model.fit(
                X,
                Y,
                epochs=self.epochs,
                batch_size=self.batch_size,
                sample_weight=sample_weight,
                shuffle=self.shuffle,
                verbose=0,
            )
        else:
            print("FS model running...")
            Y = np.expand_dims(T,-1)
            input = Input(X.shape[1:])
            output = Dense(
                Y.shape[-1] * 10,
                activation="relu",
                kernel_regularizer=SparseGroupLasso(groups=self.groups),
            )(input)
            output = Dense(1, activation=CustomSigmoid(t=1))(output)
            fs_model = Model(input, output)
            fs_model.compile(optimizer=Adam(self.learning_rate), loss=BCE, weighted_metrics=[])
            fs_model.fit(
                X,
                Y,
                epochs=self.epochs,
                batch_size=self.batch_size,
                sample_weight=sample_weight,
                shuffle=self.shuffle,
                verbose=0,
            )
        weights = fs_model.layers[1].get_weights()[0]
        weights = np.nan_to_num(weights, nan=0)
        weights = np.sum(np.square(weights), 1)
        return np.argsort(-weights),weights

    def train(self, X, y):
        seed_tensorflow()
        pos_weight = float(len(y) - y.sum()) / y.sum()
        sample_weight = np.ones(len(y))
        sample_weight[y.astype(bool)] = pos_weight
        input = Input(X.shape[1:])
        input = Dropout(0.1)(input)
        output = Dense(2, activation="relu")(input)
        output = Dense(1)(output)
        output = CustomSigmoid(t=1)(output)
        model = Model(input, output)
        model.compile(
            optimizer=Adam(self.learning_rate), loss=MSE, weighted_metrics=[]
        )
        model.fit(
            X,
            y,
            epochs=self.epochs,
            batch_size=self.batch_size,
            sample_weight=sample_weight,
            shuffle=self.shuffle,
            verbose=0,
        )
        return model

    def get_corr(self, v1: list, v2: list):
        return np.array([np.corrcoef(v1, v2[i])[0, 1] for i in range(len(v2))])

    def sort_by_group(self, vec):
        vec = self.get_corr(vec, self.data_ad.X)
        self.index = np.argsort((vec.min() - vec))
        self.data_ad = self.data_ad[self.index]
        self.groups = [i // self.group_size for i in range(len(self.index))]

    def run(self):
        geneset = pd.read_csv(self.args.input, header=None)[0]
        genes = self.data_ad.var_names.values
        y = np.in1d(genes, geneset).astype(int)
        self.sort_by_group(y)
        X = self.data_ad.X.T
        true_index = np.where(y.astype(bool))[0]
        false_index = np.random.choice(
            np.where(~y.astype(bool))[0], self.args.background_genes, replace=False
        )
        index = np.append(true_index, false_index)
        feature_indexs,weights = self.TSFS(X[index], y[index])
        feature_indexs = feature_indexs[: self.top]
        info_samples = pd.DataFrame(weights[feature_indexs], index=self.data_ad.obs_names.values[feature_indexs])
        features = X[:, feature_indexs]
        model = self.train(features[index], y[index])
        gene_vec = model.predict(features)
        gene_vec = pd.DataFrame(gene_vec, index=self.data_ad.var_names)
        return gene_vec,info_samples



import os
import re
from glob import glob
from multiprocessing import Pool
import anndata
import numpy as np
import pandas as pd
from scipy import sparse
from tqdm import tqdm

def dhs2gene(args,sample):
    try:
        r = 100000
        m = 0.01
        alpha = np.log(2 / m - 0.99) / (r - 10000)
        sample_name = re.findall(r'dhs@(.*?).csv', sample)[0]
        vec = pd.read_csv(sample, sep='\t', header=None)[0].values
        rp_vec = []
        for gene in args.genes:
            index = args.dhs_gene_g.loc[gene, 'index']
            s = vec[index]
            d = args.dhs_gene_g.loc[gene, 'DISTANCE']
            w = np.ones(d.shape)
            w[d > 10000] = 3 / (2 * np.exp((d[d > 10000] - 10000) * alpha) + 1)
            w[d > r] = 0
            rp = np.mean(np.multiply(w, s))
            rp_vec.append(rp)
        rp_vec = np.log(np.array(rp_vec) + 1)
        rp_vec = sparse.csr_matrix(rp_vec)
        return sample_name, rp_vec
    except:
        print('Error %s !' % sample)
        return None

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--library", type=str, default="library")
    parser.add_argument("--output", type=str, default="library")
    parser.add_argument("--type", type=str, default="H3K27ac",help="H3K27ac/ATAC")
    parser.add_argument("--hg38_refseq", type=str, default="library/hg38_refseq.ucsc")
    parser.add_argument("--dhs_hg38_rose", type=str, default="library/dhs_hg38_rose.bed")
    parser.add_argument("--DHS_TO_GENE", type=str, default="library/dhs_hg38_rose_DHS_TO_GENE.txt")
    args = parser.parse_args()
    assert args.type in ['H3K27ac', 'ATAC']
    
    ucsc_gene = pd.read_csv(args.hg38_refseq, sep='\t')
    ucsc_gene = ucsc_gene[['name', 'name2']]
    ucsc_gene.columns = ['GENES', 'SYMBOLS']
    ucsc_gene = ucsc_gene.drop_duplicates()
    dhs_hg38 = pd.read_csv(args.dhs_hg38_rose, sep='\t', header=None)
    dhs_hg38 = dhs_hg38.reset_index()[['index', 0]]
    dhs_hg38.columns = ['index', 'DHS']
    dhs_gene = pd.read_csv(args.DHS_TO_GENE, sep='\t')
    dhs_gene = dhs_gene.iloc[:, [0, 4, 5]]
    dhs_gene.columns = ['DHS', 'GENES', 'DISTANCE']
    dhs_gene_merge = dhs_gene.merge(dhs_hg38, on='DHS')
    dhs_gene_merge = dhs_gene_merge.merge(ucsc_gene, on='GENES')
    dhs_gene_merge = dhs_gene_merge.drop_duplicates(['DHS', 'SYMBOLS'])
    args.dhs_gene_g = dhs_gene_merge.groupby('SYMBOLS').agg(
        {'DISTANCE': list, 'index': list}
    )
    args.dhs_gene_g['DISTANCE'] = args.dhs_gene_g['DISTANCE'].apply(
        lambda x: np.array(x)
    )
    args.genes = args.dhs_gene_g.index
    rp_matrix = []
    samples = sorted(
        glob(os.path.join(args.library, f'{args.type}_matrix', 'dhs@*_hg38.csv'))
    )
    params = list(map(lambda sample: (sample), samples))
    rp_matrix = []
    sample_list = []
    with Pool(16) as pool:
        for row in tqdm(pool.imap(args.dhs2gene, params), total=len(params)):
            if row:
                sample_name, rp_vec = row
                sample_list.append(sample_name)
                rp_matrix.append(rp_vec)
    rp_matrix = sparse.vstack(rp_matrix, dtype='float32')
    rp_matrix_ad = anndata.AnnData(rp_matrix)
    rp_matrix_ad.var_names = args.genes
    rp_matrix_ad.obs_names = sample_list
    rp_matrix_ad.write_h5ad(
        os.path.join(args.output, f'RP_Matrix_{args.type}.h5ad')
    )



import os
from multiprocessing import Pool
import anndata as ad
import numpy as np
import pandas as pd
from scipy import sparse
from tqdm import tqdm

def dhs2gene(params):
    args,sample,vec = params
    try:
        r = args.range
        if args.mean:
            sample_name = sample.split("@")[0]
        else:
            sample_name = sample
        e = args.TR_info.loc[sample_name,'Distal Intergenic Percentage']
        if e > 0.01:
            m = e
        else:
            m = 0.01
        alpha = np.log(2 / m - 0.99) / (r - 10000)
        print("e: %s, alpha: %s, sample: %s" % (e,alpha,sample))
        rp_vec = []
        for gene in args.genes:
            index = args.dhs_gene_g.loc[gene,'index']
            s = vec[index]
            d = args.dhs_gene_g.loc[gene,'DISTANCE']
            w = np.ones(d.shape)
            w[d > 10000] = 3 / (2 * np.exp((d[d > 10000] - 10000) * alpha) + 1)
            w[d > r] = 0
            if args.norm:
                w_e = 3 / (2 * np.exp((r - 10000) * alpha) + 1)
                w = (w - w_e) / (1 - w_e)
            rp = np.mean(np.multiply(w,s))
            rp_vec.append(rp)
        rp_vec = sparse.csr_matrix(rp_vec)
        return sample,rp_vec
    except:
        print('Error %s !' % sample)
        return None

def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--library", type=str, default="library")
    parser.add_argument("--output", type=str, default="library")
    parser.add_argument("--hg38_refseq", type=str, default="library/hg38_refseq.ucsc")
    parser.add_argument("--dhs_hg38_rose", type=str, default="library/dhs_hg38_rose.bed")
    parser.add_argument("--DHS_TO_GENE", type=str, default="library/dhs_hg38_rose_DHS_TO_GENE.txt")
    parser.add_argument("--range", type=int, default=100000)
    parser.add_argument("--mean", type=str2bool, default=True)
    parser.add_argument("--norm", type=str2bool, default=True)
    args = parser.parse_args()

    ucsc_gene = pd.read_csv(args.hg38_refseq,sep='\t')
    ucsc_gene = ucsc_gene[['name','name2']]
    ucsc_gene.columns = ['GENES','SYMBOLS']
    ucsc_gene = ucsc_gene.drop_duplicates()
    dhs_hg38 = pd.read_csv(args.dhs_hg38_rose,sep='\t',header=None)
    dhs_hg38 = dhs_hg38.reset_index()[['index',0]]
    dhs_hg38.columns = ['index','DHS']
    dhs_gene = pd.read_csv(args.DHS_TO_GENE,sep='\t')
    dhs_gene = dhs_gene.iloc[:,[0,4,5]]
    dhs_gene.columns = ['DHS','GENES','DISTANCE']
    dhs_gene_merge = dhs_gene.merge(dhs_hg38,on='DHS')
    dhs_gene_merge = dhs_gene_merge.merge(ucsc_gene,on='GENES')
    dhs_gene_merge = dhs_gene_merge.drop_duplicates(['DHS','SYMBOLS'])
    args.dhs_gene_g = dhs_gene_merge.groupby('SYMBOLS').agg({
        'DISTANCE':list,
        'index': list
    })
    args.dhs_gene_g['DISTANCE'] = args.dhs_gene_g['DISTANCE'].apply(lambda x:np.array(x))
    args.genes = args.dhs_gene_g.index
    args.TR_info = pd.read_csv(os.path.join(args.library,'TRs_info.txt'),sep='\t',index_col=0)
    if args.mean:
        args.TR_info = args.TR_info.groupby("tr_base").agg({"Distal Intergenic Percentage":np.mean})
    tr_dhs_ad = ad.read_h5ad(os.path.join(args.library,'TR_DHS.h5ad'))

    def iter_params():
        trunk_size = 100
        trunk_count = int(tr_dhs_ad.shape[0] / trunk_size) + 1
        for trunk in tqdm(range(trunk_count)):
            start = trunk * trunk_size
            end = (trunk + 1) * trunk_size
            tr_rp = tr_dhs_ad[start:end].to_df().values
            obs = tr_dhs_ad.obs.iloc[start:end]
            for i in range(tr_rp.shape[0]):
                tr_vec = tr_rp[i]
                sample_name = obs.index[i]
                yield args, sample_name, tr_vec
    
    rp_matrix = []
    sample_list = []
    with Pool(32) as pool:
        for row in tqdm(pool.imap(dhs2gene, iter_params()),total=tr_dhs_ad.shape[0]):
            if row:
                sample_name,rp_vec = row
                sample_list.append(sample_name)
                rp_matrix.append(rp_vec)

    rp_matrix = sparse.vstack(rp_matrix,dtype='float32')
    rp_matrix_ad = ad.AnnData(rp_matrix)
    rp_matrix_ad.var_names = args.genes
    rp_matrix_ad.obs_names = sample_list

    obs = rp_matrix_ad.obs
    obs.index.name = 'tr'
    obs = obs.join(args.TR_info,how='left')
    obs['index'] = range(len(obs))

    rp_matrix_ad.obs = obs
    rp_matrix_ad.write_h5ad(os.path.join(args.output,f'RP_Matrix_TR.h5ad'))